/*
    FEARLUS model0-7: AssocArray.m
    Copyright (C) 1999-2003  Macaulay Institute

    This file is part of FEARLUS model0-7, an agent-based model of land use
    change.

    FEARLUS model0-7 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the AssocArray object.

*/

#import "AssocArray.h"
#import "Tuple.h"
#import "MiscFunc.h"

@implementation AssocArray 

/*

create: OR createBegin: ... setSize ... createEnd

Provide both options for creating the array. This enables people to override
the default size (997) during creation if they want to.

*/

+create: aZone {
  AssocArray *obj;
  int i;

  obj = [super create: aZone];
  obj->assocZone = [Zone create: aZone];
  obj->keys = [List create: obj->assocZone];
  obj->values = [List create: obj->assocZone];
  obj->size = 997;
  obj->hash = [Array create: obj->assocZone setCount: obj->size];

  for(i = 0; i < obj->size; i++) {
    [obj->hash atOffset: i put: [List create: obj->assocZone]];
  }

  return obj;
}

+createBegin: aZone {
  AssocArray *obj;

  obj = [super createBegin: aZone];
  obj->assocZone = [Zone create: aZone];
  obj->size = 997;

  return obj;
}

-setSize: (unsigned)s {
  size = s;
  while(![MiscFunc isPrime: size]) {
    size++;
  }
  return self;
}

-createEnd {
  int i;

  keys = [List create: assocZone];
  values = [List create: assocZone];
  hash = [Array create: assocZone setCount: size];

  for(i = 0; i < size; i++) {
    [hash atOffset: i put: [List create: assocZone]];
  }

  return [super createEnd];
}

/*

getKeys

Return the list of keys

*/

-(id <List>)getKeys {
  return keys;
}

/*

getValues

Return the list of values

*/

-(id <List>)getValues {
  return values;
}

/*

addObject:withKey:

Add an object to the hash table with the specified key. Return NO if the key
was not added (because the hash table doesn't like duplicates.

*/

-(BOOL)addObject: (id)anObject withKey: (id)aKeyObject {
  unsigned key;

  key = [self getKey: aKeyObject];
  if([[hash atOffset: (int)key] getCount] == 0 ||
     [self getObjectWithKey: aKeyObject] == nil) {
    Tuple *t;

    t = [Tuple create: assocZone setAlpha: aKeyObject beta: anObject];
    [[hash atOffset: (int)key] addLast: t];
    [keys addLast: aKeyObject];
    [values addLast: anObject];

    return YES;
  }
  else {
    return NO;
  }
}

/*

getObjectWithKey:

Return the object at the specified key, or nil if there wasn't any.

*/

-(id)getObjectWithKey: (id)aKeyObject {
  Tuple *t;

  t = [self getTupleWithKey: aKeyObject];
  if(t == nil) {
    return nil;
  }
  else {
    return [t getBeta];
  }
}

/*

removeKey:

If the key passed in as argument exists in the associative array, then return
the value and delete both the key and its associated value from the array.
Otherwise (if it does not exist), return nil.

*/

-(id)removeKey: (id)aKeyObject {
  Tuple *t;
  unsigned key;
  id value;

  key = [self getKey: aKeyObject];
  t = [self getTupleWithKey: aKeyObject];
  if(t != nil) {
    value = [t getBeta];
    [[hash atOffset: (int)key] remove: t];
    [keys remove: aKeyObject];
    [values remove: value];
    [t drop];
    return value;
  }
  else {
    return nil;
  }
}

/*

getTupleWithKey:

Return the Tuple stored for the key object, or nil if it wasn't found.

*/

-(Tuple *)getTupleWithKey: (id)aKeyObject {
  id inx;
  unsigned key;

  key = [self getKey: aKeyObject];

  /* Deal with the easy cases first for the sake of speed */
  if([[hash atOffset: (int)key] getCount] == 0) {
    /*				No entries, so it's a duff key */
    return nil;
  }
  else if([[hash atOffset: (int)key] getCount] == 1) {
    /*				One entry. Either it's the key or not */
    Tuple *t;

    t = [[hash atOffset: (int)key] atOffset: 0];
    if([t getAlpha] == aKeyObject) {
      return t;			// We've found the tuple
    }
    else {
      return nil;		// We haven't found it
    }
  }
  /* OK, lots of things hash to the same point. So we must check each of the
     members of the list at that hash point to see if it's the one */
  for(inx = [[hash atOffset: (int)key] begin: scratchZone], [inx next];
      [inx getLoc] == Member;
      [inx next]) {
    Tuple *t;

    t = [inx get];

    if([t getAlpha] == aKeyObject) {
      /*			We've found the tuple */
      [inx drop];
      return t;
    }
  }
  /*				The object wasn't in the list, so we haven't
				found the tuple */
  [inx drop];
  return nil;
}

/*

drop

Drop everything created by this object. Easily done because there's a separate
zone created for it, and so dropping the zone should free up all the memory.

*/

-(void)drop {
  [assocZone drop];
  [super drop];
}

/*

getKey:

Return the key for the object. To do this, we convert the address into an
unsigned long, then take the remainder of that when divided by the size of
the hash table. The conversion is done very nicely using a union.

*/

-(unsigned)getKey: (id)anObject {
  id2key converter;

  converter.obj = anObject;
  return (unsigned)(converter.key % (unsigned long)size);
}

@end
